import json
import os
import sys
from gcloud import pubsub
from gcloud import resource_manager
from google.cloud.storage.client import Client
from google.cloud.storage.bucket import Bucket
from google.oauth2 import service_account
from googleapiclient import discovery

GS_SERVICE_ACCOUNT="serviceAccount:service-{project_no}@gs-project-accounts.iam.gserviceaccount.com"
TOPIC_ID = 'mvisioncloud-pubsub-topic-event-staging-{project_no}'
SUBSCRIPTION_ID = 'mvisioncloud-pubsub-subscription-event-staging-{project_no}'
TOPIC = 'projects/{project_id}/topics/{topic}'
SUBSCRIPTION = 'projects/{project_id}/subscriptions/{sub}'


def set_gcloud_environment():
        try:
                scriptName = os.path.basename(__file__)
                for args in (sys.argv):
                        if scriptName in args:
                                dirPath = args.rstrip(scriptName)
                gcp_auth_file = dirPath + 'gcp_auth.json'
                os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = gcp_auth_file
                return gcp_auth_file
        except Exception as err:
                print(err)
                exit()


def get_storage_client():
        try:
                #storage_client = storage.client.Client()
                storage_client = Client()
                return storage_client
        except Exception as err:
                print(err)
                exit()


def get_resource_manager_client():
        try:
                rm_client = resource_manager.client.Client()
                return rm_client
        except Exception as err:
                print(err)
                exit()


def get_pubsub_client():
        pubsub_client = pubsub.client.Client()
        return pubsub_client


def get_project_id(gcp_auth_file):
        try:
                with open(gcp_auth_file) as json_file:
                        data = json.load(json_file)
                        projectId = data.get("project_id")
                        return projectId
        except Exception as err:
                print(err)
                exit()


def get_project_number(cloud_rm_object,projectId):
        try:
                projectObj = cloud_rm_object.fetch_project(projectId)
                projectNo = projectObj.number
                return projectNo
        except Exception as err:
                print(err)
                exit()

def create_pubsub_topic(pubsub_client,projectId,projectNo):
        try:
                topic_id = TOPIC_ID.format(project_no=projectNo)
                topic = TOPIC.format(project_id=projectId, topic=topic_id)
                topic_list_details = pubsub_client.publisher_api.list_topics(projectId)
                topic_list = [each_topic['name'] for each_topic in topic_list_details[0]]
                if topic not in topic_list:
                        pubsub_client.publisher_api.topic_create(topic)
                return topic_id
        except Exception as err:
                print(err)
                exit()


def create_pubsub_subscrption(pubsub_client,projectId,projectNo,topic_id):
        try:
                subscription_id = SUBSCRIPTION_ID.format(project_no=projectNo)
                subscription_name = SUBSCRIPTION.format(project_id=projectId, sub=subscription_id)
                topic = TOPIC.format(project_id=projectId, topic=topic_id)
                subsc_list_details = pubsub_client.publisher_api.topic_list_subscriptions(topic)
                if subscription_name not in subsc_list_details[0]:
                        subscription = pubsub_client.subscriber_api.subscription_create(subscription_name, topic,
                                                                                         ack_deadline=180)
        except Exception as err:
                print(err)
                exit()


def create_service_account_user(projectId,projectNo):
        try:
                credentials = service_account.Credentials.from_service_account_file(
                        filename=os.environ['GOOGLE_APPLICATION_CREDENTIALS'],
                        scopes=['https://www.googleapis.com/auth/cloud-platform'])
                iam_service = discovery.build('cloudresourcemanager', 'v1', credentials=credentials)
                policy = (
                        iam_service.projects()
                                .getIamPolicy(
                                resource=projectId,
                                body={"options": {"requestedPolicyVersion": 1}},
                        )
                                .execute()
                )

                gs_service_acc = GS_SERVICE_ACCOUNT.format(project_no=projectNo)
                try:
                        binding = next(b for b in policy["bindings"] if b["role"] == "roles/pubsub.publisher")
                        binding["members"].append(gs_service_acc)
                except StopIteration:
                        pubsub_pub_binding = {"members": [gs_service_acc], "role": "roles/pubsub.publisher"}
                        policy["bindings"].append(pubsub_pub_binding)

                policy = (
                        iam_service.projects()
                                .setIamPolicy(resource=projectId, body={"policy": policy})
                                .execute()
                )
        except Exception as err:
                print(err)
                exit()


def create_notification_for_all_buckets(storage_client,topic_id):
        try:
                buckets = storage_client.list_buckets()
                list_buckets_name = [buck.name for buck in buckets]
                for buck_name in list_buckets_name:
                        try:
                                buck_obj = Bucket(storage_client, name=buck_name)
                                nf_inst = buck_obj.list_notifications()
                                ret = buck_obj.notification(topic_id, event_types=['OBJECT_FINALIZE'],
                                                    payload_format='JSON_API_V1')
                                ret.create()
                                print("added notification for {bucket}".format(bucket=buck_name))
                        except Exception as err:
                                print(err)
        except Exception as err:
                print(err)
                exit()


def check_for_prerequisites():
        print("Prerequisites for execution of the script.")
        print("1. Please create a service account by giving project owner role and generate private key \
                                                                                        with json format")
        print("2. Copy the content of the generated json and update the same in gcp_auth.json")
        print("3. Install all the libraries [gcloud, google-api-python-client, google-cloud-storage]  if not installed before. \n")
        input_ch = input("Whether all the Prerequisites are satisfied before script execution? (Y/N) :    ")
        if input_ch in ["Y","YES","Yes","yes", "y"]:
                return True
        else:
                return False

if __name__=='__main__':
        if not(check_for_prerequisites()):
                print("Exiting... Prerequisites are not satisfied.")
                exit()
        gcp_auth_file = set_gcloud_environment()
        crm_obj = get_resource_manager_client()
        pubsub_obj = get_pubsub_client()
        project_id = get_project_id(gcp_auth_file)
        project_no = get_project_number(crm_obj, project_id)
        topic_id = create_pubsub_topic(pubsub_obj, project_id, project_no)
        create_pubsub_subscrption(pubsub_obj, project_id, project_no,topic_id)
        create_service_account_user(project_id, project_no)
        storage_obj = get_storage_client()
        create_notification_for_all_buckets(storage_obj, topic_id)
